/*
 * $Id: capinfo.c,v 1.11 1998/09/13 09:23:54 vons Exp $
 *
 * This code looks up a given ttyname in the termcap/terminfo database
 */

#include "config.h"

#include <ctype.h>
#include <string.h>

#include "vtt.h"

#ifdef HAVE_TERMCAP_H
#include <termcap.h>
#else
/* Only prototype what we actually use */
extern    PROTOTYPE(char *tgetstr, (const char *, char **));
extern    PROTOTYPE(int tgetent, (char *, const char *));
extern    PROTOTYPE(int tputs, (const char *, int, int (*)(int)));

#endif	/* HAVE_TERMCAP_H */

#ifndef HAVE_TERMCAP_VARS
/* Next var's are not used by vtt, but linking against termcap/terminfo
 * may fail without them !
 */
char      PC;
char     *UP;
char     *BC;
short     ospeed;
#endif	/* !HAVE_TERMCAP_VARS */

#ifdef HAVE_TERMCAP_IN_ENV
PROTOTYPE(static void write_termcap_entry, (char *bp));
#endif

static PROTOTYPE(int send_char, (int));

static char tcap_buf[1024];

enum CI_result
ci_test(tty_name)
const char *tty_name;
{
    enum CI_result retval;

    retval = CI_invalid;          /* make lint happy */

    switch (tgetent(tcap_buf, tty_name))
    {
      case 1:
        retval = CI_ok;
        break;
      case 0:
        retval = CI_invalid;
        break;
      default:
        pr_error(PR_FATAL, "Error opening termcap/terminfo database\n");
        break;
    }
    return retval;
}

void
ci_setup_env(tty_name)
const char *tty_name;
{
    const char *p;

    for (p = globopt.setenv_str; *p != 0; p++)
    {
        if (*p != '$')
            (void) putchar(*p);
        else
        {
            p++;
            if (*p == 'n')        /* $n = terminal-name */
                (void) fputs(tty_name, stdout);
            else if (*p == 'c')   /* $c = termcap */
            {

                /*
                 * single-quote the entry, avoid as much shell-processing
                 * as possible.
                 *
                 * Just output a null-string if there is no termcap/terminfo
                 * entry for this terminal, or $TERMCAP is not supported.
                 */
                (void) putchar('\'');
#ifdef HAVE_TERMCAP_IN_ENV
		/*
		 * Only output termcap variable if the terminal type is
		 * supported.
		 */
    		if (ci_test(tty_name) == CI_ok)
                    write_termcap_entry(tcap_buf);
#endif
                (void) putchar('\'');
            }
            else
            {
                (void) putchar('$');
                if (*p != '\0')
                    (void) putchar(*p);
                else
                    break;
            }
        }
    }

    p = globopt.setenv_str;

    if (p[strlen(p) - 1] != '\n')
        (void) putchar('\n');
}

/*
 * Output terminal initialisation string.
 */

void
ci_send_init_string()
{
    const char *init_str;
    char capabuf[256], *cbp = &capabuf[0];

    init_str = tgetstr("is", &cbp);

    if (init_str != NULL)
	(void) tputs(init_str, 1, send_char);
}

/*
 * Initialise erase key
 */
void
ci_init_tty_keys()
{
    char capabuf[256], *cbp = &capabuf[0];
    const char *bs_key;

    bs_key = tgetstr("kb", &cbp);

    if (bs_key != (const char *) NULL && strlen(bs_key) == 1)
	tty_set_erase_key(bs_key[0]);
}

/*
 * send_char -- send char to terminal.
 *
 * NOTE: this function is used by tputs() to send data to the terminal, and
 * thus must return int instead of being void
 */
static int
send_char(c)
int c;
{
    const Uchar chartosend = (const Uchar) c;

    tty_write(&chartosend, (size_t) 1);
    return 0;
}


#ifdef HAVE_TERMCAP_IN_ENV
/*
 * Output termcap entry to stdout, quoting characters that would give the
 * shell problems and omitting empty fields.
 */

static void
write_termcap_entry(bp)
char     *bp;
{
    char     *capa_start;
    char     *tok_start;
    char     *tok_end;
    const char *sep;

    /*
     * First colon indicates end of name(s) and description(s)
     */
    capa_start = strchr(bp, ':');
    if (capa_start == NULL)
        pr_error(PR_FATAL, "Invalid termcap entry\n");

    *(capa_start++) = '\0';

    /* Output all the terminal names and aliases */
    tok_start = bp;
    sep = "";

    for (;;)
    {
        tok_end = strchr(tok_start, '|');

	/* terminate with \0 if needed */
	if (tok_end != NULL)
            *tok_end = '\0';

        /* skip names containing whitespace */
        if (strchr(tok_start, ' ') == NULL && strchr(tok_start, '\t') == NULL)
        {
            (void) printf("%s%s", sep, tok_start);
            sep = "|";
        }

        if (tok_end == NULL)
            break;

        tok_start = tok_end + 1;
    }

    /* if sep didn't change, we didn't output a tty name */
    if (*sep == '\0')
        pr_error(PR_FATAL, "Invalid termcap entry\n");

    (void) putchar(':');

    tok_start = capa_start;

    for (;;)
    {
        tok_end = strchr(tok_start, ':');

        if (tok_end != NULL)
            *tok_end = '\0';

        /* skip any whitespace */
        while (*tok_start == ' ' || *tok_start == '\t')
            tok_start++;

        /*
         * Output the token char by char, escape-ing the ones that may
         * mess up the string (quotes) or have a special meaning to some
         * shell ('!', ...)
         */
        if (*tok_start != '\0')
        {
            while (*tok_start != '\0')
            {
                switch (*tok_start)
                {

                  case '"':      /* Avoid all kinds of quotes        */
                  case '\'':
                  case '`':

                  case '!':      /* Avoid csh history chars          */

                  case '*':      /* Avoid wildcards. Leaving them in */
                  case '?':      /* would make 'echo $TERMCAP' fail  */
                    		 /* with "echo: No match." ...       */

                  case '~':      /* Avoid tilde-expansion as well    */


                  case ',':      /* Berkeley code avoids comma too.  */
                                 /* Dunno why ...                    */

                    (void) printf("\\%03o", (int) *tok_start);
                    break;
                  default:
                    (void) putchar(*tok_start);
                    break;
                }
                tok_start++;
            }

            (void) putchar(':');
        }


        if (tok_end == NULL)
            break;

        tok_start = tok_end + 1;
    }
}
#endif
