/*
 * $Id: main.c,v 1.12 1998/01/11 12:35:56 vons Exp $
 */

#include "config.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <stdlib.h>
#include <signal.h>

#include "vtt.h"

#include "getopt.h"

PROTOTYPE(static void usage, (void));
PROTOTYPE(static void handle_int, (int sig));
PROTOTYPE(int main, (int argc, char **argv));

/* Declare global option structure */
struct opts globopt;

/*-------------------------------------------------------------------------
 * usage() - print usage info and exit.
 */
static void
usage()
{
    globopt.verbose = 1;          /* no problem since we exit anyway */

    pr_verbose(
      "Usage: %s [-diIKquvVx] [-e fmtstr] [-f termtab-file] [-t wait-time]\n\n",
       getprocessname());

    pr_verbose("\t-d        dump keyword and query info (implies -v)\n");
    pr_verbose("\t-e fmtstr output TERM and TERMCAP definitions\n");
#ifndef HAVE_TERMCAP_IN_ENV
    pr_verbose("\t          (NOTE: $c is not available on this system!)\n");
#endif
    pr_verbose("\t-f file   specifies the termtab_file to use\n");
    pr_verbose("\t-i        ignore users private .vtt_termtab\n");
    pr_verbose("\t-I        don't send initialisation string to terminal\n");
    pr_verbose("\t-K        don't set the terminal's erase key\n");
    pr_verbose("\t-q        quiet\n");
    pr_verbose("\t-t n      set wait timeout to n/10 seconds, 1 <= n <= 100\n");
    pr_verbose("\t-u        don't try the entries for unsupported terminals\n");
    pr_verbose("\t-v        verbose\n");
    pr_verbose("\t-V        print version\n");
    pr_verbose("\t-x        trace processing of .vtt_termtab\n");

    exit(2);
}

/*-------------------------------------------------------------------------
 * main() ...
 */

int
main(argc, argv)
int       argc;
char    **argv;
{

    int       opt;
    char     *ttytype;
    char     *tty_description;
    int       found;
    const char *termtab_name;

    setprocessname(argv[0]);

    globopt.dump_queries = 0;
    globopt.trace = 0;
    globopt.verbose = 0;
    globopt.quiet = 0;
    globopt.skip_unknown_ttys = 0;
    globopt.setenv_str = NULL;
    globopt.no_init = 0;
    globopt.dont_set_keys = 0;
    termtab_name = NULL;

    opterr = 0;
    while ((opt = getopt(argc, argv, ":de:f:iIKquvVt:x")) != EOF)
    {
        switch (opt)
        {
          case 'd':
            globopt.dump_queries = 1;
            globopt.verbose = 1;
            break;
          case 'e':
            globopt.setenv_str = optarg;
            break;
          case 'f':
            termtab_name = optarg;
            break;
          case 'i':
            tt_ignore_usertab();
            break;
	  case 'I':
	    globopt.no_init = 1;
	    break;
	  case 'K':
	    globopt.dont_set_keys = 1;
	    break;
          case 'q':
            globopt.quiet = 1;
            break;
          case 't':
            {
                long      wto = atol(optarg);

                if (wto < 1 || wto > 100)
                {
                    pr_error(PR_WARN, "Invalid wait-timeout %li\n", wto);
                    usage();
                }

                query_settimeout((int) wto);
            }
            break;
          case 'u':
            globopt.skip_unknown_ttys = 1;
            break;
          case 'v':
            globopt.verbose = 1;
            break;
          case 'V':
            pr_error(PR_WARN, "Version 1.1.2\n");
            break;
          case 'x':
            globopt.trace = 1;
            break;
          case ':':
            pr_error(PR_WARN, "Option '%c' misses argument\n\n", optopt);
            usage();
            break;
          case '?':
            pr_error(PR_WARN, "Invalid option '%c'\n\n", optopt);
            usage();
            break;
          default:
            (void) abort();
        }
    }

    if (argc - optind != 0)
        pr_error(PR_WARN, "Arguments are ignored\n");

    /* Make sure we are connected to a tty */
    tty_verify();

    /*
     * Locate & open the termtab file, and read the first part of it
     */
    tt_init(termtab_name);

    /* Catch interrupts so that we can restore tty settings */
    (void) signal(SIGINT, handle_int);
    (void) signal(SIGTERM, handle_int);

    /* Set tty in raw mode */
    tty_raw();

    if (globopt.dump_queries)
        query_dump();

    /* Interrogate and report findings */
    found = (tt_process(&ttytype, &tty_description) == TT_recognised);

    /* Restore tty mode */
    tty_restore();

    if (!found)
    {

        ttytype = getenv("TERM");

        if (ttytype == NULL || ttytype[0] == '\0')
        {
            ttytype = VTT_DFLT_TERMTYPE;
            if (!globopt.quiet)
                pr_error(PR_WARN, "$TERM is not defined\n");
        }

        if (!globopt.quiet)
            pr_error(PR_WARN, "Terminal not recognised, defaulting to %s\n",
                     ttytype);

        tty_description = "<default, terminal not recognised>";
    }

    if (ci_test(ttytype) != CI_ok)
    {

	pr_verbose("Terminal is recognised as '%s' (%s)\n",
               ttytype, tty_description != NULL ? tty_description : "");

	/* force exit status != 0 */
        found = 0;

        if (!globopt.quiet)
        {
            pr_error(PR_WARN,
        "Could not find this terminal type in the termcap/terminfo database!\n"
                );
        }
    }
    else if (found)
    {
	/* This must only be done if tty type is known, _and_ supported */
	if (globopt.no_init == 0)
	    ci_send_init_string();

	/* display verbose message here, now that the cursor won't be moved
	 * to an unknown place anymore (ci_send_init_string() could do that).
	 */
	pr_verbose("Terminal is recognised as '%s' (%s)\n",
               ttytype, tty_description != NULL ? tty_description : "");

	if (globopt.dont_set_keys == 0)
	    ci_init_tty_keys();
    }

    tt_exit();

    if (globopt.setenv_str == NULL)
        (void) puts(ttytype);
    else
        ci_setup_env(ttytype);

    return (found ? 0 : 1);
}

static void
handle_int(sig)
int       sig;
{
    /* pr_error() does not return */
    pr_error(PR_FATAL, "Interrupted (signal number %i), aborted\n", sig);
}
