/*
 * $Id: parse.c,v 1.5 1996/05/19 19:41:49 vons Exp $
 *
 * Code that reads the input, returning words/strings/...
 */

#include "config.h"
#include "vtt.h"

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>


#define ESC	27


PROTOTYPE(static void skip_whitesp, (void));
PROTOTYPE(static void pa_uncook, (const char *cstr, size_t clen,
                                  Uchar ** rstr, size_t *rlen));

static char line[VTT_MAX_LINELEN + 1];
static const char *curpos;
static int lineno = 0;
static FILE *pa_fp;

void
pa_setinput(fp)
FILE     *fp;
{
    pa_fp = fp;

    /* init the module by reading a line */
    if (pa_nextline() == PA_eof)
        pr_error(PR_FATAL, "Empty .vtt_termtab file, terminating...\n");
}

int
pa_lineno()
{
    return lineno;
}

void
pa_restart_line()
{
    curpos = &line[0];
}

enum PA_result
pa_nextline()
{
    char     *p;

    while (fgets(line, sizeof(line), pa_fp) != NULL)
    {
        lineno++;

        /* Make sure line is not too long */
        if (strlen(line) == (size_t) VTT_MAX_LINELEN &&
            line[VTT_MAX_LINELEN - 1] != '\n')
        {
            pr_error(PR_FATAL, "Input line '%i' too long\n", pa_lineno());
        }

        for (p = line; *p != 0 && isspace(*p);)
            p++;

        /* line empty or contains only a comment ? */
        if (*p == 0 || *p == '#')
            continue;

        /* line starts with a non-comment, try it */
        pa_restart_line();
        return PA_ok;
    }

    /* EOF */
    return PA_eof;
}

enum PA_result
pa_getoperator(wsp, condp)
enum PA_wsp wsp;
enum PA_ops *condp;
{
    if (wsp == PA_wsp_skip)
        skip_whitesp();

    switch (*curpos)
    {
      case '=':
        *condp = PA_op_equal;
        break;
      case '#':
        *condp = PA_op_unequal;
        break;
      case 0:
        return PA_eol;
      default:
        return PA_fail;
    }

    curpos++;
    return PA_ok;
}

/*------------------------------------------------------------------------
 * pa_getline() - return remainder of line in alloced buffer.
 */
char     *
pa_getline(wsp)
enum PA_wsp wsp;
{
    const char *p;
    char     *pdup;
    size_t    len;

    if (wsp == PA_wsp_skip)
        skip_whitesp();

    p = curpos;

    while (*curpos && *curpos != '\n')
        curpos++;

    /*
     * duplicate, make sure there is a trailing nul (memdup() handles
     * that)
     */
    len = (size_t) (curpos - p);
    pdup = (char *) memdup(p, len, 1);

    return pdup;
}

enum PA_result
pa_getword(wsp, wptr)
enum PA_wsp wsp;
char    **wptr;
{
    const char *w_start;
    size_t    len;

    if (wsp == PA_wsp_skip)
        skip_whitesp();

    if (*curpos == 0)
    {

        *wptr = NULL;
        return PA_eol;
    }

    w_start = curpos;

    /* Word must start with an alpha-numeric char, or be a $something */
    if (!isalnum(*curpos) && *curpos != '$')
    {
        *wptr = NULL;
        return PA_fail;
    }

    /*
     * Find the end of the word, but pass the (potential) '$' first.
     *
     * Word may contain '_', '-'... (possible tty-name!)
     */
    curpos++;
    while (*curpos && (isalnum(*curpos) || *curpos == '-' || *curpos == '_'))
        curpos++;

    len = (size_t) (curpos - w_start);

    /*
     * Return a nul-terminated word to the caller. (the extra 1 at the end
     * is a nul-byte)
     */
    *wptr = (char *) memdup(w_start, len, 1);

    return PA_ok;
}


enum PA_result
pa_findchar(wsp, chr)
enum PA_wsp wsp;
int       chr;
{
    enum PA_result ret;

    if (wsp == PA_wsp_skip)
        skip_whitesp();

    if (*curpos == 0)
        ret = PA_eol;
    else
        ret = (*curpos == chr ? PA_ok : PA_fail);

    /* move to next char */
    curpos++;

    return ret;
}


enum PA_result
pa_getstr(wsp, sptr, slen)
enum PA_wsp wsp;
Uchar   **sptr;
size_t   *slen;
{
    enum PA_result res;
    const char *str_start;
    size_t    len;

    /* Get to first char of string */
    res = pa_findchar(wsp, '"');

    if (res != PA_ok)
        return res;

    str_start = curpos;

    /*
     * Find end of string. Note that findchar() placed curpos _after_ the
     * opening quote, so (curpos-1) is valid!
     *
     * Look for \n as well, just in case the trailing quote is missing.
     */
    while (*curpos && *curpos != '\n' &&
           (*curpos != '"' || *(curpos - 1) == '\\'))
    {
        curpos++;
    }

    len = curpos - str_start;

    if (*curpos == '"')
    {
        /* Everything ok, move to char _after_ trailing quote for next op */
        curpos++;
    }
    else
    {
        pr_error(PR_WARN, "No trailing quote on line %i\n", pa_lineno());

        /*
         * Just go on, use end of line as string end. Don't move curpos if
         * it points to the trailing 0. Otherwise, remove \n from string
         * and set curpos on the 0.
         */
        if (*curpos == '\n')
        {
            curpos++;

            /* Handle case where opening quote is last char on line */
            if (len != 0)
                len--;
        }
    }

    /* convert it to the raw string */
    pa_uncook(str_start, len, sptr, slen);

    return PA_ok;
}

static void
pa_uncook(cstr, clen, rstr, rlen)
const char *cstr;
size_t    clen;
Uchar   **rstr;
size_t   *rlen;
{
#define isoctal(c)	('0' <= (c) && (c) < '8')

    const char *pread, *pend;
    Uchar    *pwrite;
    Uchar    *buf;

    /*
     * String can only shrink, so the cooked len is more than enough. Just
     * use memdup() to get a zero-filled string.
     */
    buf = (Uchar *) memdup(cstr, 0, clen);
    pwrite = buf;
    pread = cstr;
    pend = pread + clen;

    while (pread < pend)
    {
        switch (*pread)
        {
          case '\\':
            pread++;
            if (isoctal(*pread))
            {

                /*
                 * Convert \ooo ==> char, accept max 3 octal digits.
                 */
                int       i;
                int       num = 0;

                for (i = 1; i <= 3 && isoctal(*pread); i++)
                {
                    num = num * 8 + (*pread - '0');
                    pread++;
                }

                *(pwrite++) = (Uchar) num;
            }
            else
            {
                /* No octal digit, convert \x ==> x */
                *(pwrite++) = (Uchar) * (pread++);
            }

            break;

          case '^':
            /* Convert ^X ==> cntrl-X */
            pread++;

            /* ^[ stands for an escape */
            if (*pread == '[')
            {
                *(pwrite++) = ESC;
                pread++;
            }
            else if (isalpha(*pread))
            {
                if (isupper(*pread))
                    *pwrite = (Uchar) ((*pread - 'A') + 1);
                else
                    *pwrite = (Uchar) ((*pread - 'a') + 1);

                pread++;
                pwrite++;
            }
            else
            {
                pr_error(PR_WARN, "Invalid control sequence '^%c' on line %i\n",
                         *pread, pa_lineno());

                /*
                 * Don't fail, just use the caret and leave the second
                 * char for the next round.
                 */
                *(pwrite++) = (Uchar) '^';
            }
            break;

          default:
            *pwrite = (Uchar) * pread;
            pread++;
            pwrite++;
        }                         /* end switch() */

    }                             /* end while() */

    *rstr = buf;
    *rlen = (size_t) (pwrite - buf);
}

static void
skip_whitesp()
{
    while (*curpos && isspace(*curpos))
        curpos++;
}
