/*
 * $Id: query.c,v 1.7 1998/01/11 12:30:40 vons Exp $
 *
 * This file contains the code that manages the tty query strings and
 * the other keywords.
 */

#include "config.h"

#ifdef HAVE_SYS_UTSNAME_H
#include <sys/utsname.h>          /* uname() */
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>             /* setitimer() */
#endif

#include <sys/param.h>            /* MAXHOSTNAMELEN */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#ifdef HAVE_PWD_H
#include <pwd.h>
#endif

#include "vtt.h"


struct query
{
    const char *name;
    const Uchar *esc_sequence;
    size_t    esc_seqlen;

              PROTOTYPE(void (*exec), (struct query *));

    const Uchar *answer;
    size_t    answerlen;
    struct
    {
        char      used;
        char      executed;
    }         flags;
};

PROTOTYPE(static void q_tty, (struct query *));
PROTOTYPE(static void q_hostname, (struct query *));
PROTOTYPE(static void q_osname, (struct query *));
PROTOTYPE(static void q_osversion, (struct query *));
PROTOTYPE(static void q_machine, (struct query *));
PROTOTYPE(static void q_ttyname, (struct query *));
PROTOTYPE(static void q_termvar, (struct query *));

PROTOTYPE(static int install_timeout_handler, (void));
PROTOTYPE(static void timeout_handler, (int sig));
PROTOTYPE(static int settimer, (int timeout));
#ifdef HAVE_UNAME
PROTOTYPE(static struct utsname * call_uname, (void));
#endif

/*
 * Next table contains the predefined keywords, the query sting,
 * the function will obtain the response, and the reply.
 *
 * The keywords _must_ be uppercase, the code depends on it!
 */
static struct query qtab[] =
{
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},  /* 1 */
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},  /* 10 */
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},
    {"", NULL, 0, q_tty, NULL, 0, {0, 0}},  /* 20 */

    {"HOSTNAME", NULL, 0, q_hostname, NULL, 0, {1, 0}},
    {"OSNAME", NULL, 0, q_osname, NULL, 0, {1, 0}},
    {"OSVERSION", NULL, 0, q_osversion, NULL, 0, {1, 0}},
    {"MACHINE", NULL, 0, q_machine, NULL, 0, {1, 0}},
    {"TTYNAME", NULL, 0, q_ttyname, NULL, 0, {1, 0}},
    {"TERM", NULL, 0, q_termvar, NULL, 0, {1, 0}},
    {NULL, NULL, 0, NULL, NULL, 0, {0, 0}}
};

static int timeout_value = VTT_DFLT_TIMEOUT;
static int timeout_seen;

/*----------------------------------------------------------------
 * query_settimeout() - set the wait-timeout to given value
 */
void
query_settimeout(t)
int       t;
{
    timeout_value = t;
}

/*----------------------------------------------------------------
 * query() - lookup the keyword in the list, and return its value.
 *
 *           If the keyword is a query which is referenced for the
 *	     first time, the query-string is sent to the terminal
 *	     and the answer is stored as the value of the keyword.
 */
void
query(qid, answer, answerlen)
const char *qid;
const Uchar **answer;
size_t   *answerlen;
{
    struct query *q;

    q = &qtab[0];

    /* scan the table */
    while (q->name != NULL && strcmp(q->name, qid) != 0)
        q++;

    /*
     * If keyword doesn't exist or the query is undefined in the
     * vtt_termtab, the caller should know about it...
     */
    if (q->name == NULL || q->flags.used == 0)
    {
        *answer = NULL;
        *answerlen = 0;
        return;
    }

    /*
     * If it is the first time, execute the query and store it for future
     * reference.
     */
    if (q->flags.executed == 0)
    {
        q->exec(q);
        q->flags.executed = 1;
    }

    /*
     * Return the result
     */
    *answer = q->answer;
    *answerlen = q->answerlen;
}

/*----------------------------------------------------------------
 * query_add() - Find an empty slot, and add the query to the table
 */
void
query_add(qid, qstr, qlen)
const char *qid;
const Uchar *qstr;
size_t    qlen;
{
    struct query *q, *first_empty;

    /* Look up the query, remember the first empty slot */
    for (q = &qtab[0], first_empty = NULL; q->name != NULL; q++)
    {
        /* slot already in use ? */
        if (q->flags.used == 1)
        {

            /*
             * Redefinition is currently not allowed, it might indicate an
             * error.
             */
            if (strcmp(q->name, qid) == 0)
            {
                pr_error(PR_FATAL, "Redefinition of query '%s' not allowed\n",
                         qid);
            }
        }
        else if (first_empty == NULL)
        {
            /* First empty slot found, remember it */
            first_empty = q;
        }
    }

    if (first_empty == NULL)
    {
        pr_error(PR_FATAL, "Can't store query '%s', table is full\n", qid);
    }

    first_empty->name = qid;
    first_empty->flags.used = 1;

    first_empty->esc_sequence = qstr;
    first_empty->esc_seqlen = qlen;
    first_empty->flags.executed = 0;
}

/*----------------------------------------------------------------
 * q_tty() - send the esc-sequence to the terminal and wait for
 *	     the response.
 */
static void
q_tty(q)
struct query *q;
{
    Uchar     recbuf[100];
    size_t    read_cnt;

    assert(q->flags.executed == 0);

    if (!install_timeout_handler())
    {
        pr_error(PR_FATAL, "Could not install signal handler, %s\n",
                 strerror(errno));
    }

    tty_write(q->esc_sequence, q->esc_seqlen);

    timeout_seen = 0;

    if (!settimer(timeout_value))
    {
        pr_error(PR_FATAL, "Could not start timer\n");
    }

    read_cnt = 0;

    /*
     * Read on until timeout arrives, can't stop earlier than that since
     * length of reply is unknown!
     *
     * qterm also stops if the character 'c' is received since a lot of
     * replies have a trailing 'c'. Don't do that here since it is not
     * 100% reliable, and the gain is minimal since vtt supports
     * sub-second timeouts anyway.
     */
    while (read_cnt < sizeof(recbuf) && !timeout_seen)
    {
        int       read_result;

        read_result = tty_readchar();

        /*
         * read_result == -1 means EINTR, either due to timeout, or other
         * intr. just restart loop and let the guard check for this.
         */
        if (read_result == -1)
            continue;

        recbuf[read_cnt++] = (Uchar) read_result;
    }

    /* store reply */
    q->answerlen = read_cnt;
    q->answer = (const Uchar *) memdup(recbuf, q->answerlen, 0);

}

/*----------------------------------------------------------------
 * timeout_handler() - handles the ALRM signal, just sets a flag
 */
static void
/*ARGSUSED*/
timeout_handler(sig)
int       sig;
{
    timeout_seen = 1;
}

/*----------------------------------------------------------------
 * q_hostname() - return the name of the host. Supports both
 *                uname() and gethostname().
 */
static void
q_hostname(q)
struct query *q;
{
    char     *hn;

    hn = NULL;
#ifdef HAVE_UNAME
    hn = call_uname()->nodename;
#else
#ifdef HAVE_GETHOSTNAME
    {
        static char name[MAXHOSTNAMELEN];

        if (gethostname(name, MAXHOSTNAMELEN) == -1)
        {
            pr_error(PR_FATAL, "gethostname() failed, %s\n", strerror(errno));
        }

        hn = name;
    }
#else
    hn = "unknown";
#endif
#endif
    q->answerlen = strlen(hn);
    q->answer = (const Uchar *) hn;
}

/*----------------------------------------------------------------
 * q_osname() -  return the name of the Operating System.
 */
static void
q_osname(q)
struct query *q;
{
    char     *on;
#ifdef HAVE_UNAME

    on = call_uname()->sysname;
#else
    on = "unknown";
#endif
    q->answerlen = strlen(on);
    q->answer = (const Uchar *) on;
}

/*----------------------------------------------------------------
 * q_osversion() -  return the version of the Operating System.
 */
static void
q_osversion(q)
struct query *q;
{
    char     *ov;
#ifdef HAVE_UNAME
    ov = call_uname()->release;
#else
    ov = "unknown";
#endif
    q->answerlen = strlen(ov);
    q->answer = (const Uchar *) ov;
}

/*----------------------------------------------------------------
 * q_machine() -  return the type of the machine.
 */
static void
q_machine(q)
struct query *q;
{
    char     *ma;
#ifdef HAVE_UNAME
    ma = call_uname()->machine;
#else
    ma = "unknown";
#endif
    q->answerlen = strlen(ma);
    q->answer = (const Uchar *) ma;
}

/*----------------------------------------------------------------
 * q_ttyname() - return the name of the tty device.
 */
static void
q_ttyname(q)
struct query *q;
{
    char     *ttynam;
    ttynam = ttyname(STDIN_FILENO);

    if (ttynam == NULL)
    {
        pr_error(PR_FATAL, "ttyname() failed, %s\n", strerror(errno));
    }

    q->answerlen = strlen(ttynam);
    q->answer = (const Uchar *) ttynam;
}

/*----------------------------------------------------------------
 * q_termvar() - returns the value of the $TERM environment variable
 */
static void
q_termvar(q)
struct query *q;
{
    const char *termvar = getenv("TERM");

    if (termvar == NULL || termvar[0] == '\0')
    {
        termvar = "unknown";

        pr_trace("$TERM undefined, using '%s' instead\n", termvar);
    }

    /* copy str + trailing nul */
    q->answerlen = strlen(termvar);
    q->answer = (const Uchar *) memdup(termvar, strlen(termvar), 1);
}

/*----------------------------------------------------------------
 * call_uname() - calls uname() only once and stores the result
 *		  for later use.
 */
#ifdef HAVE_UNAME
static struct utsname *
call_uname()
{
    static int uname_called = 0;
    static struct utsname name;

    if (!uname_called)
    {
        if (uname(&name) == -1)
        {
            pr_error(PR_FATAL, "uname() failed, %s\n", strerror(errno));
        }

        uname_called = 1;
    }

    return &name;
}
#endif

/*----------------------------------------------------------------
 * settimer() - start a real-time timer for the given interval.
 *
 * Returns true if successful
 *
 *
 * The code below supports
 *
 *	- setitimer() (POSIX)
 *	- ularm()     (BSD)
 *      - alarm()     (ANSI-C) value is rounded to the nearest second.
 */

#ifdef HAVE_SETITIMER

#define TIMER_OK
#undef HAVE_UALARM

#include <sys/time.h>

static int
settimer(timeout)
int       timeout;
{
    struct itimerval ival;

    ival.it_interval.tv_sec = 0;
    ival.it_interval.tv_usec = 0;
    ival.it_value.tv_sec = timeout / 10;
    ival.it_value.tv_usec = 100000 * (timeout % 10);

    return (setitimer(ITIMER_REAL, &ival, NULL) == 0);
}

#endif

#ifdef HAVE_UALARM
#define TIMER_OK

static int
settimer(timeout)
int       timeout;
{

    /*
     * ualarm() cannot return an error, it only returns the time left on
     * the clock before the change. Should be 0 here: there is no repeat,
     * and the program waits for the expiration every time.
     *
     * NOTE: ualarm() wants microseconds.
     */
    return (ualarm(100000 * timeout, 0) == 0);
}
#endif

#ifndef TIMER_OK

#include <stdlib.h>
static int
settimer(timeout)
int       timeout;
{

    /*
     * alarm() cannot return an error, it only returns the time left on
     * the clock before the change. Should be 0 here: there is no repeat,
     * and the program waits for the expiration every time.
     *
     * NOTE: alarm() wants seconds.
     */
    return (alarm((unsigned int) ((timeout + 9) / 10)) == 0);
}
#endif

/*----------------------------------------------------------------
 * install_timeout_handler() - install a handler for the ALRM signal
 *
 * Returns true if the install failed.
 *
 * This code supports
 *	- sigaction() - POSIX
 *	- signal()    - ANSI. Just pray that this interrupts a read()...
 */
#ifdef HAVE_SIGACTION
static int
install_timeout_handler()
{
    struct sigaction act;

    act.sa_handler = timeout_handler;
    (void) sigemptyset(&act.sa_mask);

    act.sa_flags = 0;

#ifdef SA_RESTART
    /* sa_flags=0, no SA_RESTART */
#else
#ifdef SA_INTERRUPT
    act.sa_flags = SA_INTERRUPT;
#else
    /* leave sa_flags=0, and hope for the best ... */
#endif
#endif

    return (sigaction(SIGALRM, &act, NULL) == 0);
}
#else                             /* fall back to signal() */

static int
install_timeout_handler()
{
    return (signal(SIGALRM, timeout_handler) != SIG_ERR);
}
#endif

/*----------------------------------------------------------------
 * query_dump() - display all defined keywds, and their value
 */
void
query_dump()
{
    struct query *q;
    const Uchar *ans;
    size_t    anslen;

    pr_verbose("%-20s   %-10s   %s\n", "Keyword", "Query", "Value");
    pr_verbose("----------------------------------------------\n");

    for (q = &qtab[0]; q->name != NULL; q++)
    {
	if(q->flags.used == 0)
	    continue;

        /* Keyword */
        pr_verbose("%-20s   ", q->name);

        query(q->name, &ans, &anslen);

        /* Query */

        if (ans == NULL)
        {
            /* pr_verbose("<unused>\n"); */
	    ;
        }
        else if (q->esc_sequence == NULL)
        {
            pr_verbose("%-10s   ", "");
        }
        else
        {
            char     *qstr;

            qstr = tt_cook(q->esc_sequence, q->esc_seqlen);

            pr_verbose("%-10s   ", qstr);

            (void) free(qstr);
        }

        /* Value */

        if (ans == NULL)
        {
            /* Query <undefined> so no value to report */
        }
        else if (anslen == 0)
        {
            pr_verbose("<empty string>\n");
        }
        else
        {
            char     *val;

            val = tt_cook(ans, anslen);

            pr_verbose("%s\n", val);

            (void) free(val);
        }

    }
}
