/*
 * $Id: support.c,v 1.4 1995/11/05 17:16:51 vons Exp $
 *
 * Various support funtions.
 */
#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#ifdef HAVE_WORKING_STDARG
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include "vtt.h"

static const char *glb_prgname = "undefined";

/*--------------------------------------------------------------------------
 * Error handling is done with a varargs function. Support both
 * the ANSI stdarg.h as well as the older varargs.h
 */
#ifdef HAVE_WORKING_STDARG
void
pr_error(int act, const char *fmtstr,...)
{
    va_list   ap;
    va_start(ap, fmtstr);
#else
void
pr_error(va_alist)
va_dcl
{
    va_list   ap;
    char     *fmtstr;
    int       act;

    va_start(ap);
    act = va_arg(ap, int);
    fmtstr = va_arg(ap, char *);
#endif

    tty_restore();
    (void) fprintf(stderr, "%s: ", glb_prgname);
    (void) vfprintf(stderr, fmtstr, ap);

    (void) fflush(stderr);

    if (act == PR_FATAL)
        exit(2);

    tty_raw();

    va_end(ap);
}


#ifdef HAVE_WORKING_STDARG
void
pr_verbose(const char *fmtstr,...)
{
    va_list   ap;
    va_start(ap, fmtstr);
#else
void
pr_verbose(va_alist)
va_dcl
{
    va_list   ap;
    char     *fmtstr;

    va_start(ap);
    fmtstr = va_arg(ap, char *);
#endif

    if (globopt.verbose == 1)
    {
        tty_restore();
        (void) vfprintf(stderr, fmtstr, ap);
        tty_raw();
    }

    va_end(ap);
}

#ifdef HAVE_WORKING_STDARG
void
pr_trace(const char *fmtstr,...)
{
    va_list   ap;
    va_start(ap, fmtstr);
#else
void
pr_trace(va_alist)
va_dcl
{
    va_list   ap;
    char     *fmtstr;

    va_start(ap);
    fmtstr = va_arg(ap, char *);
#endif

    if (globopt.trace == 1)
    {
        tty_restore();
        (void) vfprintf(stderr, fmtstr, ap);
        tty_raw();
    }

    va_end(ap);
}


/*--------------------------------------------------------------------------
 * setprocessname - get basename of argument string and store it for later.
 */
void
setprocessname(argv0)
const char *argv0;
{
    const char *pbasename, *pext;
    char     *p;

    /* some systems don't have a valid argv[0] */
    if (0 == argv0)
        argv0 = "vtt";

    pbasename = strrchr(argv0, '\\'); /* OS/2 and MSDOS allow \ */

    if (pbasename == (char *) 0)
        pbasename = strrchr(argv0, '/');  /* others are normal */

    if (pbasename == (char *) 0)
        pbasename = argv0;
    else
        pbasename++;              /* skip '/' or '\' */

    /*
     * Set pext to first char after basename, in order to strip possible
     * extension.
     */
    pext = strrchr(pbasename, '.');
    if (pext == (char *) 0)
        pext = strchr(pbasename, '\0');

    p = (char *) malloc((size_t) (pext - pbasename + 1));

    if (p == NULL)                /* malloc() failed, bad luck */
        glb_prgname = pbasename;

    (void) strncpy(p, pbasename, (size_t) (pext - pbasename));

    p[pext - pbasename] = '\0';

    glb_prgname = p;
}

const char *
getprocessname()
{
    return glb_prgname;
}

/*--------------------------------------------------------------------------
 * memdup - like strdup(), but does not rely on trailing 0-byte, and alloc's
 *          additional trailing bytes on request.
 */
void     *
memdup(src, len, extra)
const void *src;
size_t    len;
size_t    extra;
{
    char     *dst;

    /* avoid problems with malloc(), always alloc at least 1 byte */
    if (len + extra == 0)
        extra = 1;

    dst = (char *) malloc(len + extra);

    if (dst == NULL)
        pr_error(PR_FATAL, "Not enough memory...\n");

    (void) memcpy(dst, src, len);
    (void) memset(dst + len, 0, extra);

    return (void *) dst;
}

/*--------------------------------------------------------------------------
 * strtoupper - make whole string uppercase.
 */
void
strtoupper(str)
char     *str;
{
    while (*str)
    {
        if (islower(*str))
            *str = (char) toupper(*str);

        str++;
    }
}

#ifndef HAVE_STRERROR
/*--------------------------------------------------------------------------
 *
 * strerror() function for those who don't have it. Given an errno value,
 * it returns a description of the error.
 */

#include <stdio.h>
#include <errno.h>

extern int sys_nerr;
extern char *sys_errlist[];

static char msg[80];

char     *
strerror(errnum)
int       errnum;
{

    if ((errnum >= 0) && (errnum < sys_nerr))
        return sys_errlist[errnum];

    (void) sprintf(msg, "Unknown errno value %i\n", errnum);
    return msg;
}
#endif
