/*
 * $Id: termtab.c,v 1.7 1996/05/19 19:42:24 vons Exp $
 *
 * This module locates the vtt config file, and evaluates its contents.
 */

#include "config.h"
#include "vtt.h"

#include <sys/types.h>
#include <sys/stat.h>

#ifdef HAVE_PWD_H
#include <pwd.h>
#endif

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#define ESC 27

static const char dflt_termtab_name[] = VTT_USER_TERMTAB;
static const char dflt_system_termtab[] = VTT_SYSTEM_TERMTAB;

static int check_usertab = 1;
static int do_skip;
static FILE *termtab_fp;

PROTOTYPE(static void tt_read_queries, (void));
PROTOTYPE(static void tt_parse_ttytype,
          (char *wptr, char **ttytype, char **ttydesc));
PROTOTYPE(static enum TT_result tt_eval_line, (void));
PROTOTYPE(static int tt_strcmp, (const Uchar * s1, const Uchar * s2,
                                 size_t len1, size_t len2));

void
tt_ignore_usertab()
{
    check_usertab = 0;
}

void
tt_init(forced_termtab_name)
const char *forced_termtab_name;
{
    char     *homedir = NULL;
    char     *termtab_file;

    if (forced_termtab_name != NULL)
    {
        pr_trace("Trying specified termtab '%s'...", forced_termtab_name);

        termtab_fp = fopen(forced_termtab_name, "r");

        if (termtab_fp != NULL)
            goto tabfile_ok;

        pr_trace("failed\n");
        pr_error(PR_FATAL, "Cannot open '%s'\n", forced_termtab_name);
    }

#ifdef HAVE_PWD_H
    {
        struct passwd *pwent;

        pwent = getpwuid(getuid());

        if (pwent != NULL)
            homedir = pwent->pw_dir;
        else
            pr_error(PR_FATAL, "You (uid %li) don't exist!\n", (long) getuid());
    }
#endif

    if (homedir == NULL && getenv("HOME") != NULL)
        homedir = getenv("HOME");

    if (homedir == NULL)
        pr_error(PR_FATAL, "Can't find your home directory\n");

    termtab_file = malloc(strlen(homedir) + strlen(dflt_termtab_name) + 2);

    if (termtab_file == NULL)
        pr_error(PR_FATAL, "Not enough memory...\n");

    (void) strcpy(termtab_file, homedir);
    (void) strcat(termtab_file, "/");
    (void) strcat(termtab_file, dflt_termtab_name);

    if (check_usertab == 0)
    {
        pr_trace("Ignoring user termtab\n");
        termtab_fp = NULL;
    }
    else
    {
        pr_trace("Trying user termtab '%s'...", termtab_file);

        termtab_fp = fopen(termtab_file, "r");

        if (termtab_fp != NULL)
            goto tabfile_ok;

        pr_trace("failed\n");
    }

    if (termtab_fp == NULL)
    {
        pr_trace("Trying system termtab '%s'...", dflt_system_termtab);
        termtab_fp = fopen(dflt_system_termtab, "r");

        if (termtab_fp != NULL)
            goto tabfile_ok;

        pr_trace("failed\n");
        pr_error(PR_FATAL, "No readable termtab file found\n");
    }

tabfile_ok:

    pr_trace("ok\n");

    pa_setinput(termtab_fp);

    pr_trace("Reading queries...\n");

    tt_read_queries();
}

void
tt_exit()
{
    (void) fclose(termtab_fp);
}

static void
tt_read_queries()
{

    char     *wptr;
    Uchar    *sptr;
    char     *qid;
    size_t    slen;
    enum PA_result res;

    /* Loop until a line is found that doesn't contain a 'query = ...' */

    for (;;)
    {
        res = pa_getword(PA_wsp_fail, &wptr);
        if (res == PA_ok)
            res = pa_findchar(PA_wsp_skip, '=');

        /*
         * Assume it is the first ttydescr line if there is no valid word
         * or no equal sign,
         */
        if (res != PA_ok)
        {
            pr_trace("Line %i is first line with tty description\n",
                     pa_lineno());

            /* reset line before further processing */
            pa_restart_line();

            return;
        }

        /* copy the qid, ... */
        qid = memdup(wptr, strlen(wptr), 1);

        /* ... get the string ... */
        if (pa_getstr(PA_wsp_skip, &sptr, &slen) != PA_ok)
        {
            pr_error(PR_WARN, "No opening quote found on line %i\n",
                     pa_lineno());
            goto next_line;
        }

        /* ... and add it to the list of queries */

        /* lookup and definition of queries assumes uppercase */
        strtoupper(qid);

        query_add(qid, sptr, slen);

        /*
         * We are done with this line (ignore remainder of the line)
         */
next_line:
        if (pa_nextline() == PA_eof)
            pr_error(PR_FATAL, "No tty description found\n");

    }

}

enum TT_result
tt_process(ttytype, ttydesc)
char    **ttytype;
char    **ttydesc;
{
    int       expr_seen;
    char     *wptr;

    *ttytype = NULL;
    *ttydesc = NULL;
    expr_seen = 0;
    do_skip = 0;                  /* skipmode off */

    for (;;)
    {
        enum PA_result res;

        /* get first word of the ttydesc line, should be a tty name */
        res = pa_getword(PA_wsp_fail, &wptr);

        /* The line is non-empty by definition */
        assert(res != PA_eol);

        if (res == PA_ok)
        {

            /*
             * This is a ttytype line. See if there was a condition line
             * before it, unless this is the first entry.
             */
            if (!expr_seen && *ttytype != NULL)
            {
                pr_error(PR_WARN,
                         "No condition for ttytype '%s' near line %i\n",
                         *ttytype, pa_lineno());
            }

            expr_seen = 0;
            tt_parse_ttytype(wptr, ttytype, ttydesc);
        }
        else
        {

            /*
             * getword failed, so first char was not alnum(). It may be a
             * condition line, so try again, but look for the '>' this
             * time.
             */
            pa_restart_line();
            res = pa_findchar(PA_wsp_fail, '>');

            /* hmmm, trouble... report the error, but ignore it */
            if (res != PA_ok)
            {
                pr_error(PR_WARN, "Invalid line near line %i\n", pa_lineno());
            }
            else
            {
                expr_seen = 1;

                if (!do_skip)
                {
                    if (tt_eval_line() == TT_recognised)
                        return TT_recognised;
                }
            }
        }

        if (pa_nextline() == PA_eof)
            break;

    }                             /* for(;;) */

    return TT_failed;
}

static void
tt_parse_ttytype(wptr, ttytype, ttydesc)
char     *wptr;
char    **ttytype;
char    **ttydesc;
{
    enum PA_result res;

    /* Get the ttytype */

    if (*ttytype != NULL)
        (void) free(*ttytype);

    if (*ttydesc != NULL)
        (void) free(*ttydesc);

    /*
     * a $something refers to keyword 'something'
     */
    if (wptr[0] != '$')
        *ttytype = wptr;
    else
    {
        const Uchar *ans;
        size_t    anslen;

        /* keyword lookup assumes upper-case */
        strtoupper(wptr);

        query(&(wptr[1]), &ans, &anslen);

        if (ans != NULL)
        {
            (void) free(wptr);
            /* assume that the retrieved value makes sense ... */
            *ttytype = (char *) memdup(ans, anslen, 1);
        }
        else
        {
            pr_error(PR_WARN, "Invalid keyword reference '%s' on line %i\n",
                     wptr, pa_lineno());

            /* Don't bail out, just go on with a '$something' */
            *ttytype = wptr;
        }
    }


    /* Get the tty description */

    res = pa_findchar(PA_wsp_skip, '|');

    if (res == PA_ok)
        *ttydesc = pa_getline(PA_wsp_skip);
    else if (res == PA_eol)
    {
        static const char nodesc[] = "<no description available>";

        /* must alloc, a free() will follow! */
        *ttydesc = (char *) memdup(nodesc, sizeof(nodesc), 1);
    }
    else
    {
        pr_error(PR_WARN, "Unrecognised data after tty name '%s' on line %i\n",
                 *ttytype, pa_lineno());
    }


    if (globopt.skip_unknown_ttys == 0 || ci_test(*ttytype) == CI_ok)
    {
        pr_trace("\nLine %i: Trying '%s' (%s) \n",
                 pa_lineno(), *ttytype,
                 (*ttydesc != NULL ? *ttydesc : "<none>"));

        do_skip = 0;
    }
    else
    {
        pr_trace("\nLine %i: Skipping '%s' (%s) [Not supported]\n",
                 pa_lineno(), *ttytype,
                 (*ttydesc != NULL ? *ttydesc : "<none>"));

        pr_verbose("terminal type '%s' (line %i) not supported, skipping... \n",
                   *ttytype, pa_lineno());

        do_skip = 1;
    }
}

/*
 * Evaluate a 'condition line'.
 *
 * Syntax:
 *
 * <cline>     ::= <leader> <expression>
 *
 * <leader>    ::= '>'
 *
 * <expression>::= <condition>
 *		 | <condition> '&' <expression>
 *
 * <condition> ::= <keyword> <op> <string>
 *
 * <op>	       ::= '=' | '#'
 *
 * <keyword>   ::= {one of the keywords defined in query.c}
 *
 * <string>    ::= <quote> <charseq> <quote>
 *
 * <quote>     ::= '"'
 *
 * <charseq>   ::= <ch> | 0ddd | "\"<ch>
 *
 */
static enum TT_result
tt_eval_line()
{

    char     *keyw = NULL;
    Uchar    *string = NULL;
    const Uchar *ans;
    size_t    string_len;
    size_t    anslen;
    enum PA_result res;

    /* process everything up to eol */
    while ((res = pa_getword(PA_wsp_skip, &keyw)) != PA_eol)
    {
        enum PA_ops op_type;
        int       fail;

        if (res == PA_fail)
        {
            pr_error(PR_WARN, "Invalid or missing keyword on line %i\n", pa_lineno());
            goto tt_eval_end;
        }

        /* lookup the value of the symbol */
        strtoupper(keyw);
        query(keyw, &ans, &anslen);

        if (ans == NULL)
        {
            pr_error(PR_WARN, "Invalid keyword '%s' on line %i\n",
                     keyw, pa_lineno());
            goto tt_eval_end;
        }

        if (pa_getoperator(PA_wsp_skip, &op_type) != PA_ok)
        {
            pr_error(PR_WARN, "Invalid or missing operator on line %i\n",
                     pa_lineno());
            goto tt_eval_end;
        }

        if (pa_getstr(PA_wsp_skip, &string, &string_len) != PA_ok)
        {
            pr_error(PR_WARN, "Invalid or missing string operand on line %i\n",
                     pa_lineno());
            goto tt_eval_end;
        }

        if (op_type == PA_op_equal)
            fail = !tt_strcmp(ans, string, anslen, string_len);
        else
            fail = tt_strcmp(ans, string, anslen, string_len);

        if (globopt.trace)
        {
            char     *pans, *pstr;

            pans = tt_cook(ans, anslen);

            pstr = tt_cook(string, string_len);

            pr_trace("Line %i: > '%s'(%s) %c '%s': %s\n",
                     pa_lineno(), keyw,
                     *pans ? pans : "<empty string>",
                     op_type == PA_op_equal ? '=' : '#',
                     pstr,
                     fail ? "Failed" : "Ok"
                );

            (void) free(pans);
            (void) free(pstr);
        }

        (void) free(string);
        string = NULL;

        (void) free(keyw);
        keyw = NULL;

        if (fail)
            goto tt_eval_end;

        /* Find '&' sym that separates exprs */
        if ((res = pa_findchar(PA_wsp_skip, '&')) != PA_ok)
        {
            if (res == PA_eol)
            {

                /*
                 * This line evaluated to true!
                 */
                return TT_recognised;
            }
            else
            {
                pr_error(PR_WARN, "Missing '&' on line %i\n", pa_lineno());
                goto tt_eval_end;
            }
        }
        /* Re-start for next condition on this line */
    }

tt_eval_end:
    if (keyw != NULL)
        (void) free(keyw);
    if (string != NULL)
        (void) free(string);

    /* return failure... */
    return TT_failed;
}

static int
tt_strcmp(s1, s2, len1, len2)
const Uchar *s1;
const Uchar *s2;
size_t    len1, len2;
{
    if (len1 < len2)
    {
        return 0;
    }
    else if (len2 == 0)
    {
        /* only successful if len1 is 0 as well */
        return (len1 == 0);
    }
    else
    {
        /* len1 >= len2 > 0 */

        /* if s2 ends in wildcard, compare all but the last char */
        if (s2[len2 - 1] == '*')
        {
            return (memcmp(s1, s2, len2 - 1) == 0);
        }
        else
        {
            /* if they have same length, compare whole string */
            return (len1 == len2 && memcmp(s1, s2, len2) == 0);
        }
    }
}

char     *
tt_cook(s, len)
const unsigned char *s;
size_t    len;
{

    char     *cook, *p;
    int       i;

    cook = (char *) malloc(4 * len + 1);  /* 4* is max expansion */

    if (cook == NULL)
    {
        pr_error(PR_FATAL, "Malloc failure...");
    }

    p = cook;
    for (i = 0; i < len; i++)
    {
        if (s[i] <= 26)
        {
            *p++ = '^';
            *p++ = (char)( 'A' + s[i] - 1 );
        }
        else if (s[i] == ESC)
        {
            *p++ = '^';
            *p++ = '[';
        }
        else if (isprint((int) s[i]))
        {
            *p++ = s[i];
        }
        else
        {
            *p++ = '\\';
            *p++ = (char)( '0' + ((s[i] >> 6) & 0x03) );
            *p++ = (char)( '0' + ((s[i] >> 3) & 0x07) );
            *p++ = (char)( '0' + ((s[i] >> 0) & 0x07) );
        }
    }

    *p = 0;

    return cook;
}
