/*
 * $Id: tty.c,v 1.9 1998/01/11 12:26:43 vons Exp $
 *
 * This module handles the tty-modes, both setting it to raw-mode, and
 * restoring it on request. The code supports the termios, termio, and
 * sgttyb interfaces.
 */

#include <sys/types.h>            /* possibly ssize_t */
#include <sys/ioctl.h>
#include <unistd.h>

#include <string.h>
#include <errno.h>

#include "config.h"
#include "vtt.h"

#ifndef STDIN_FILENO
#define STDIN_FILENO	0
#endif

#ifndef STDERR_FILENO
#define STDERR_FILENO	2
#endif

enum TTY_result
{
    TTY_ok, TTY_getfailed, TTY_setfailed, TTY_flushfailed
};

PROTOTYPE(static enum TTY_result ttymode_drain, (void));
PROTOTYPE(static enum TTY_result ttymode_set_raw, (void));
PROTOTYPE(static enum TTY_result ttymode_restore, (void));
PROTOTYPE(static enum TTY_result ttymode_set_erase_key, (int key));

static int tty_read_fd = STDIN_FILENO;
static int tty_write_fd = STDERR_FILENO;

/* tty_mode >= 0 normal, tty_mode < 0 is raw */
static int tty_mode = 0;

/*----------------------------------------------------------------------
 * tty_raw() - Set tty to raw mode.
 */
void
tty_raw()
{
    /* Change only from 0 to -1 */
    if (tty_mode != 0)
    {
        tty_mode--;
        return;
    }

    switch (ttymode_set_raw())
    {
      case TTY_ok:
        break;
      case TTY_getfailed:
        pr_error(PR_FATAL, "Could not get tty settings, %s\n", strerror(errno));
        break;
      case TTY_setfailed:
        pr_error(PR_FATAL, "Could not set raw mode, %s\n", strerror(errno));
        break;
      default:
        break;
    }

    tty_mode--;
}

/*----------------------------------------------------------------------
 * tty_restore() - Restore original ttymode
 */
void
tty_restore()
{
    if (tty_mode == -1)
    {
        if (ttymode_restore() != TTY_ok)
            pr_error(PR_FATAL, "Could not restore tty mode, %s\n",
                     strerror(errno));
    }

    tty_mode++;
}

/*----------------------------------------------------------------------
 * tty_verify() - makes sure that the standard input really is a terminal.
 */
void
tty_verify()
{
    if (!isatty(STDIN_FILENO))
    {
        pr_error(PR_FATAL, "Standard input is not a terminal!\n");
    }
}

/*----------------------------------------------------------------------
 * tty_readchar() - read a char from the tty, return -1 when interrupted.
 */
int
tty_readchar()
{
    char      c;
    int       retval;

    retval = -1;                  /* to satisfy lint ... */

    switch (read(tty_read_fd, &c, 1))
    {
      case 1:
        retval = (unsigned char) c; /* make sure 8bit values are positive! */
        break;
      case 0:
        pr_error(PR_FATAL, "read returned EOF ?!?!?\n");
        /* NOTREACHED */
        break;
      default:
        if (errno == EINTR)
            retval = -1;
        else
        {
            pr_error(PR_FATAL, "could not read from terminal, %s\n",
                     strerror(errno));
        }
    }

    return retval;
}

/*----------------------------------------------------------------------
 * tty_write() - write string to tty and wait for output to drain.
 */
void
tty_write(buf, len)
const Uchar *buf;
size_t    len;
{
    ssize_t   write_res;

    write_res = write(tty_write_fd, (const void *) buf, len);

    /*
     * Don't check for and loop on EINTR, not worth it...
     */
    if (write_res != (ssize_t) len)
    {
        tty_restore();
        pr_error(PR_FATAL, "could not write to screen, %s\n",
                 (write_res < 0 ? strerror(errno) : "reason unknown"));
    }

    /* wait until output has been written */
    if (ttymode_drain() == TTY_flushfailed)
    {
	/* EINVAL may indicate that the flush ioctl is not supported,
	 * e.g. under UnixWare 2.1.2.
	 */
	if (errno != EINVAL)
	    pr_error(PR_FATAL, "Flush failed, %s\n", strerror(errno));
    }
}

void
tty_set_erase_key(key)
int key;
{
    if (ttymode_set_erase_key(key) != TTY_ok)
	pr_error(PR_WARN, "Could not set erase key\n");
    else
    {
	char keyname[3];
	keyname[1] = keyname[2] = '\0';

	if (key < 32)					/* control key */
	    keyname[0] = '^', keyname[1] = (char) ('@' + key);
	else if (key == 127)				/* DEL key */
	    keyname[0] = '^', keyname[1] = '?';
	else						/* other */
	    keyname[0] = (char) key;

	pr_verbose("Erase key set to '%s'\n", keyname);
    }
}

/*----------------------------------------------------------------------------
 * termios version of ttymode_set_raw(), ttymode_restore(), and ttymode_drain()
 */
#ifdef HAVE_TERMIOS_H
#define TTY_OK

#include <termios.h>

static struct termios tty_orig;

static enum TTY_result
ttymode_set_raw()
{
    struct termios tty_new;

    if (tcgetattr(tty_write_fd, &tty_orig) == -1)
        return TTY_getfailed;

    tty_new = tty_orig;

    tty_new.c_iflag &= ~(ISTRIP | INLCR | IGNCR | ICRNL | IXON);
    tty_new.c_oflag &= ~(OPOST);
    tty_new.c_lflag &= ~(ICANON | ISIG | ECHO);
    tty_new.c_cc[VMIN] = 1;
    tty_new.c_cc[VTIME] = 0;

    /* put terminal in raw mode, flush any pending data */
    if (tcsetattr(tty_write_fd, TCSAFLUSH, &tty_new) == -1)
        return TTY_setfailed;

    return TTY_ok;
}

static enum TTY_result
ttymode_restore()
{
    if (tcsetattr(tty_write_fd, TCSAFLUSH, &tty_orig) == -1)
        return TTY_setfailed;

    return TTY_ok;
}

static enum TTY_result
ttymode_drain()
{
    if (tcdrain(tty_write_fd) == -1)
        return TTY_flushfailed;

    return TTY_ok;
}

static enum TTY_result
ttymode_set_erase_key(key)
int key;
{
    struct termios tty;

    if (tcgetattr(tty_write_fd, &tty) == -1)
        return TTY_getfailed;

    tty.c_cc[VERASE] = (char) key;

    if (tcsetattr(tty_write_fd, TCSAFLUSH, &tty) == -1)
        return TTY_setfailed;

    return TTY_ok;
}

#endif                            /* HAVE_TERMIOS_H */

/*----------------------------------------------------------------------------
 * termio version of ttymode_set_raw(), ttymode_restore(), and ttymode_drain()
 */

#ifndef TTY_OK
#ifdef HAVE_TERMIO_H
#define TTY_OK

#include <termio.h>

static struct termio tty_orig;

static enum TTY_result
ttymode_set_raw()
{
    struct termio tty_new;

    if (ioctl(tty_write_fd, TCGETA, &tty_orig) == -1)
        return TTY_getfailed;

    tty_new = tty_orig;

    tty_new.c_iflag &= ~(ISTRIP | INLCR | IGNCR | ICRNL | IXON);
    tty_new.c_oflag &= ~(OPOST);
    tty_new.c_lflag &= ~(ICANON | ISIG | ECHO);
    tty_new.c_cc[VMIN] = 1;
    tty_new.c_cc[VTIME] = 0;

    /* put terminal in raw mode, flush any pending data */
    if (ioctl(tty_write_fd, TCSETA, &tty_new) == -1)
        return TTY_setfailed;

    return TTY_ok;
}

static enum TTY_result
ttymode_restore()
{
    if (ioctl(tty_write_fd, TCSETA, &tty_orig) == -1)
        return TTY_setfailed;

    return TTY_ok;
}

static enum TTY_result
ttymode_drain()
{
    int       nobrk = 1;

    if (ioctl(tty_write_fd, TCSBRK, nobrk) == -1)
        return TTY_flushfailed;

    return TTY_ok;
}

static enum TTY_result
ttymode_set_erase_key(key)
int key;
{
    struct termio tty;

    if (ioctl(tty_write_fd, TCGETA, &tty) == -1)
        return TTY_getfailed;

    tty.c_cc[VERASE] = key;

    if (ioctl(tty_write_fd, TCSETA, &tty) == -1)
        return TTY_setfailed;

    return TTY_ok;
}

#endif                            /* HAVE_TERMIO_H */
#endif                            /* TTY_OK */

/*---------------------------------------------------------------------------
 * sgtty version of ttymode_set_raw(), ttymode_restore(), and ttymode_drain()
 */

#ifndef TTY_OK
#ifdef HAVE_SGTTY_H
#define TTY_OK

#include <sgtty.h>

static struct sgttyb tty_orig;

static enum TTY_result
ttymode_set_raw()
{
    struct sgttyb tty_new;

    if (ioctl(tty_write_fd, TIOCGETP, &tty_orig) == -1)
        return TTY_getfailed;

    tty_new = tty_orig;

    tty_new.sg_flags &= ~ECHO;
    tty_new.sg_flags |= RAW;

    /* put terminal in raw mode, flush any pending data */
    if (ioctl(tty_write_fd, TIOCSETP, &tty_new) == -1)
        return TTY_setfailed;

    return TTY_ok;
}

static enum TTY_result
ttymode_restore()
{
    if (ioctl(tty_write_fd, TIOCSETP, &tty_orig) == -1)
        return TTY_setfailed;

    return TTY_ok;
}

static enum TTY_result
ttymode_drain()
{
    /* no idea how to flush... */
    return TTY_ok;
}

void
ttymode_set_erase_key(key)
int key;
{
    struct sgttyb tty;

    if (ioctl(tty_write_fd, TIOCGETP, &tty == -1)
        return TTY_getfailed;

    tty.sg_erase = key;

    if (ioctl(tty_write_fd, TIOCGETP, &tty == -1)
        return TTY_getfailed;

    return TTY_ok;
}


#endif                            /* HAVE_SGTTY_H */
#endif                            /* TTY_OK */
